/*
 * Copyright © 2008 Intel Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 *
 * Authors:
 *    Eric Anholt <eric@anholt.net>
 *    Keith Packard <keithp@keithp.com>
 *
 */

#include "drmP.h"
#include "drm.h"
#include "radeon_drm.h"
#include "radeon_drv.h"

static int radeon_ring_info(char *buf, char **start, off_t offset,
			       int request, int *eof, void *data)
{
	struct drm_minor *minor = (struct drm_minor *) data;
	struct drm_device *dev = minor->dev;
	drm_radeon_private_t *dev_priv = dev->dev_private;
	uint32_t rdp, wdp;
	int len = 0;

	if (offset > DRM_PROC_LIMIT) {
		*eof = 1;
		return 0;
	}

	*start = &buf[offset];
	*eof = 0;
	rdp = RADEON_READ(RADEON_CP_RB_RPTR);
	wdp = RADEON_READ(RADEON_CP_RB_WPTR);
	DRM_PROC_PRINT("RADEON_CP_RB_WPTR %08x\n", wdp);
	DRM_PROC_PRINT("RADEON_CP_RB_RPTR %08x\n", rdp);

	if (rdp != wdp) {
		uint32_t *start, *end;
		uint32_t *ptr;
		rdp = (rdp + 15) & ~15;
		if (rdp > 128)
			rdp -= 128;
		else
			rdp = 0;
		start = dev_priv->ring.start + rdp;
		if (rdp > wdp) {
			/* dump the ring until the end */
			end = dev_priv->ring.start + dev_priv->ring.size;
		} else
			end = dev_priv->ring.start + wdp;
			
		for (ptr = start; ptr < end; ptr += 4) {
			DRM_PROC_PRINT("%x: %08x %08x %08x %08x\n", ptr - dev_priv->ring.start,
				       ptr[0], ptr[1], ptr[2], ptr[3]);
		}
	}
	
	if (len > request + offset)
		return request;
	*eof = 1;
	return len - offset;
}

static int radeon_ib_info(char *buf, char **start, off_t offset,
			  int request, int *eof, void *data)
{
	struct drm_minor *minor = (struct drm_minor *) data;
	struct drm_device *dev = minor->dev;
	drm_radeon_private_t *dev_priv = dev->dev_private;
	uint32_t ib_base, ib_size, csq_stat;
	uint32_t rdp, wdp;
	int len = 0;

	if (offset > DRM_PROC_LIMIT) {
		*eof = 1;
		return 0;
	}

	*start = &buf[offset];
	*eof = 0;
	rdp = RADEON_READ(RADEON_CP_RB_RPTR);
	wdp = RADEON_READ(RADEON_CP_RB_WPTR);

	if (rdp == wdp) {
		*eof = 1;
		return 0;
	}

	/* find the IB that is being processed */
	ib_base = RADEON_READ(RADEON_CP_IB_BASE);
	ib_size = RADEON_READ(RADEON_CP_IB_BUFSZ);
	csq_stat = RADEON_READ(RADEON_CP_CSQ_STAT);

	DRM_PROC_PRINT("RADEON_CP_IB_BASE/BUFSZ %08x/%08x\n", ib_base, ib_size);
	DRM_PROC_PRINT("RADEON_CP_CSQ_STAT %08x/indrptr %08x\n", csq_stat, csq_stat >> 20);
	if (ib_base && ib_size) {
		int i;

		for (i = 0; i < RADEON_NUM_IB; i++)
			if ((dev_priv->ib_objs[i]->bo->offset + dev_priv->gart_vm_start) == ib_base)
				break;
		if (i < RADEON_NUM_IB) {
			uint32_t offset;
			uint32_t *start, *end;
			uint32_t *ptr;
			offset = csq_stat >> 20;
			if (offset > 32)
				offset -= 32;
			else
				offset = 0;
			offset = (offset + 15) & ~15;
			start = dev_priv->ib_objs[i]->kmap.virtual + (offset * sizeof(u32));
			end = dev_priv->ib_objs[i]->kmap.virtual + (ib_size * sizeof(u32));
			for (ptr = start; ptr < end; ptr += 4) {
				DRM_PROC_PRINT("%x: %08x %08x %08x %08x\n", ptr - (uint32_t *)dev_priv->ib_objs[i]->kmap.virtual,
					       ptr[0], ptr[1], ptr[2], ptr[3]);
			}
		}
	}
	
	if (len > request + offset)
		return request;
	*eof = 1;
	return len - offset;
}

static int radeon_interrupt_info(char *buf, char **start, off_t offset,
			       int request, int *eof, void *data)
{
	struct drm_minor *minor = (struct drm_minor *) data;
	struct drm_device *dev = minor->dev;
	drm_radeon_private_t *dev_priv = dev->dev_private;
	int len = 0;

	if (offset > DRM_PROC_LIMIT) {
		*eof = 1;
		return 0;
	}

	*start = &buf[offset];
	*eof = 0;
	DRM_PROC_PRINT("Interrupt enable:    %08x\n",
		       RADEON_READ(RADEON_GEN_INT_CNTL));

	if (dev_priv->chip_family >= CHIP_RS690) {
	  DRM_PROC_PRINT("DxMODE_INT_MASK:         %08x\n",
			 RADEON_READ(R500_DxMODE_INT_MASK));
	}
	DRM_PROC_PRINT("Interrupts received: %d\n",
		       atomic_read(&dev_priv->irq_received));
	DRM_PROC_PRINT("Current sequence:    %d %d\n",
		       READ_BREADCRUMB(dev_priv), RADEON_READ(RADEON_SCRATCH_REG3));
	DRM_PROC_PRINT("Counter sequence:     %d\n",
		       dev_priv->counter);
	if (dev_priv->chip_family >= CHIP_R300) 
		DRM_PROC_PRINT("CS:    %d\n",
			       GET_SCRATCH(6));

	
	if (len > request + offset)
		return request;
	*eof = 1;
	return len - offset;
}

static struct drm_proc_list {
	/** file name */
	const char *name;
	/** proc callback*/
	int (*f) (char *, char **, off_t, int, int *, void *);
} radeon_gem_proc_list[] = {
	{"radeon_gem_interrupt", radeon_interrupt_info},
	{"radeon_gem_ring", radeon_ring_info},
	{"radeon_gem_ib", radeon_ib_info},
};


#define RADEON_GEM_PROC_ENTRIES ARRAY_SIZE(radeon_gem_proc_list)

int radeon_gem_proc_init(struct drm_minor *minor)
{
	struct proc_dir_entry *ent;
	int i, j;

	for (i = 0; i < RADEON_GEM_PROC_ENTRIES; i++) {
		ent = create_proc_entry(radeon_gem_proc_list[i].name,
					S_IFREG | S_IRUGO, minor->dev_root);
		if (!ent) {
			DRM_ERROR("Cannot create /proc/dri/.../%s\n",
				  radeon_gem_proc_list[i].name);
			for (j = 0; j < i; j++)
				remove_proc_entry(radeon_gem_proc_list[i].name,
						  minor->dev_root);
			return -1;
		}
		ent->read_proc = radeon_gem_proc_list[i].f;
		ent->data = minor;
	}
	return 0;
}

void radeon_gem_proc_cleanup(struct drm_minor *minor)
{
	int i;

	if (!minor->dev_root)
		return;

	for (i = 0; i < RADEON_GEM_PROC_ENTRIES; i++)
		remove_proc_entry(radeon_gem_proc_list[i].name, minor->dev_root);
}
